/*
This file is part of myFitter, a C++ library for maximum likelihood fits.
Copyright (C) 2012, 2013  Martin Wiebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef MYFITTER_FITTER_HPP
#define MYFITTER_FITTER_HPP

#include "model.hpp"
#include "inputfunction.hpp"
#include "minimizer.hpp"

#include <exception>
#include <boost/math/special_functions/gamma.hpp>
#include <gsl/gsl_multimin.h>
#include <dvegas/dvegas.h>


namespace myfitter {

// exception class for large Delta chi-squares

class ELargeDeltaChiSquare : public std::exception {
    virtual const char* what() const throw()
    {
        return "Delta chi-square value is too large for double precision arithmetic.";
    }
};

class EFailedModelEvaluation : public std::exception {
public:
    int status;
    EFailedModelEvaluation(int i) : status(i) { }
    virtual const char* what() const throw()
    {
        return ("model evalutation failed with exit status "
                + boost::lexical_cast<std::string, int>(status)).c_str();
    }
};


// main Fitter class

class Fitter {
public:
    class LogLikelihoodFunction : public ObjectiveFunction {
    private:
        Fitter* fitter_;
        Model* model_;
        const ObservableVector* cvals_;
        IndexVector freepar_;
        int status_;
        int verbose_;
        double chisq0_;
        
    public:
        LogLikelihoodFunction() { }
        LogLikelihoodFunction(Fitter& fitter,
                              Model& model,
                              const ObservableVector& cvals,
                              int verbosity = 0);
        LogLikelihoodFunction(const LogLikelihoodFunction& f) :
            ObjectiveFunction(f),
            fitter_(f.fitter_),
            model_(f.model_),
            cvals_(f.cvals_),
            freepar_(f.freepar_),
            status_(f.status_),
            verbose_(f.verbose_),
            chisq0_(f.chisq0_) { }
        LogLikelihoodFunction& operator=(const LogLikelihoodFunction& f) {
            ObjectiveFunction::operator=(f);
            fitter_ = f.fitter_;
            model_ = f.model_;
            cvals_ = f.cvals_;
            freepar_ = f.freepar_;
            status_ = f.status_;
            verbose_ = f.verbose_;
            chisq0_ = f.chisq0_;
            return *this;
        }

        virtual double operator()(const ParameterVector& x);
        virtual void get_gradient(const ParameterVector& x, ParameterVector& g);
        virtual void get_fdf(const ParameterVector& x,
                            double& f, ParameterVector& g);

        int nfreepar() const { return freepar_.size(); }
        int freepar(int i) const { return freepar_[i]; }
        void get_minimization_parameters(ParameterVector& x) const;
        double lower_bound() const { return chisq0_; }
    };

    // Integrand functor for p-value integrations

    class PValueIntegrand : public HepSource::Integrand {
        Model* fmod_;
        Model* cmod_;
        Fitter* fitter_;
        IndexVector claimedobs_;
        Matrix lam2obs_;
        ObservableVector origin_, lambda0_;
        int ndim0_, ndim1_, ndim_, nvar_;
        double dchisq_;
        double jacobian_;
        int nfailedshots_;
        int verbose_;
        bool adaptive_;

        void set_lambda_(const double x[], // adapted integration variables 
                                           // (uses 2 for rsqinner > 0,
                                           // 0 otherwise)
                         const double y[], // normal distributed random
                                           // variables (uses ndim)
                         int ndim,         // number of dimensions to set
                         int offset,       // start at dimension 'offset'
                         double rsqinner,  // square of radius defining
                                           // inner region
                         ObservableVector& lambda, // vector to fill
                         double& jacobian);// this gets multiplied with
                                           // the jacobian of the mapping
    public:
        PValueIntegrand(Model& fmod,       // full model
                        Model& cmod,       // constrained model
                        Fitter& fitter,    // fitter object to fit the models
                        const IndexVector& claimedobs, // claimed observables
                        const Matrix& lam2obs, // conversion from normalised
                                               // to ordinary observables
                        const ObservableVector& origin, // origin in Q-space
                        const ObservableVector& lambda0, // mod[0] best-fit
                                               // point in lambda-space
                        int n0,   // pick radius r in first n0 dimensions
                        int n1,   // pick radius in dimensions n0..n1-1,
                                  // omitting inner region defined by
                                  // r^2 + deltachisq
                        double deltachisq, // delta chi^2 value to beat
                        double jacobian,   // jacobian factor
                        int verbose,
                        bool adaptive
                        );

        virtual void operator()(const double x[], const int k[],
                                const double& weight, const double aux[],
                                double f[]);

        // return number of auxiliary dimensions
        int nvar() const { return nvar_; }

        // return number of failed shots
        int failed_shots() const { return nfailedshots_; }

        // set number of failed shots to zero
        void reset() { nfailedshots_ = 0; }
    };


    // Integrand functor for adaptive scans

    class ScanIntegrand : public HepSource::Integrand {
    private:
        Model* mod_;
        InputFunction* ifunc_;
        ObservableVector cvals_;
        bool store_;
        double chisquare_offset_;
        double chisquare_power_;

    public:
        ScanIntegrand(Model& mod, InputFunction& ifunc,
                      const ObservableVector& cvals,
                      double offset=1., double power=1.) :
            HepSource::Integrand(mod.nparameters(),1),
            mod_(&mod), ifunc_(&ifunc), cvals_(cvals),
            store_(false), chisquare_offset_(offset), chisquare_power_(power)
        { }
        ScanIntegrand(const ScanIntegrand& i) :
            HepSource::Integrand(i), mod_(i.mod_), ifunc_(i.ifunc_),
            cvals_(i.cvals_), store_(false),
            chisquare_offset_(i.chisquare_offset_),
            chisquare_power_(i.chisquare_power_) { }

        virtual const ScanIntegrand& operator=(const ScanIntegrand& i) {
            HepSource::Integrand::operator=(i);
            mod_ = i.mod_;
            ifunc_ = i.ifunc_;
            cvals_ = i.cvals_;
            store_ = i.store_;
            chisquare_offset_ = i.chisquare_offset_;
            chisquare_power_ = i.chisquare_power_;
            return *this;
        }

        bool store() const { return store_; }
        void store(bool s) { store_ = s; }

        double chisquare_offset() const { return chisquare_offset_; }
        void chisquare_offset(double d) { chisquare_offset_ = d; }

        double chisquare_power() const { return chisquare_power_; }
        void chisquare_power(double d) { chisquare_power_ = d; }

        virtual void operator()(const double x[], const int k[],
                                const double& weight, const double aux[],
                                double f[]);        
    };

private:
    int npar_, nobs_;
    InputFunction ifunc_;
    Matrix qo_, oq_;
    ObservableVector cobs_, fobs_;

    double orthogonalization_tolerance_;
    BasicMinimizer* minimizer_;
    int minimizer_verbose_;
    double minimizer_step_width_;
    double minimizer_line_search_precision_;
    double minimizer_precision_;
    int minimizer_iterations_;
    bool minimizer_release_parameters_;
    bool minimizer_keep_hessian_;
    double inner_region_fraction_;
    double inner_region_power_;

    double scans_chisquare_offset_;
    double scans_chisquare_power_;
    int scans_nbins_;

    int dvegas_verbose_;
    int dvegas_nbins_;
    int dvegas_nfirstshots_;
    int dvegas_nshots_;
    int dvegas_niterations_;
    double dvegas_precision_;
    bool dvegas_adaptive_;

    double pvalue_, pvalue_error_;
    bool reached_precision_goal_;
    double failed_shot_ratio_;
    HepSource::Int64 nshots_;

    void assign_deriv_(Matrix& m, int icol, const Model& mod, int ipar);
    void set_conversion_matrices_(const Model& fmod, const Model& cmod);
    double bfgs_test_gradient_(gsl_multimin_fdfminimizer* s,
                               ParameterVector& oldx,
                               ParameterVector& oldgrad,
                               Matrix& invhess);
    int local_fit_(Model& model, const ObservableVector& cvals);

public:
    // constructor
    Fitter(int no) :
        nobs_(no), ifunc_(no),
        orthogonalization_tolerance_(1.e-6),
        minimizer_(0),
        minimizer_verbose_(0),
        minimizer_step_width_(1.),
        minimizer_line_search_precision_(0.1),
        minimizer_precision_(1.e-3),
        minimizer_iterations_(200),
        minimizer_keep_hessian_(false),
        inner_region_fraction_(0.1),
        inner_region_power_(1.),
        scans_chisquare_offset_(1.),
        scans_chisquare_power_(1.),
        scans_nbins_(50),
        dvegas_verbose_(0),
        dvegas_nbins_(50),
        dvegas_nfirstshots_(1000),
        dvegas_nshots_(500),
        dvegas_niterations_(20),
        dvegas_precision_(0.01),
        dvegas_adaptive_(true),
        pvalue_(Model::nan()),
        pvalue_error_(Model::nan()),
        reached_precision_goal_(false),
        failed_shot_ratio_(0.)
    { minimizer(BFGSMinimizer()); }

    ~Fitter() { if(minimizer_) delete minimizer_; }

    // number of observables
    int nobservables() const { return nobs_; }
    
    // access input function
    InputFunction& input_function() { return ifunc_; }
    const InputFunction& input_function() const { return ifunc_; }


    // get/set the the tolerance used for constructing an orthonormal basis
    // from the derivatives of normalized observables with respect to
    // parameters. A (projection of) a derivative vector with respect to
    // parameter i is considered zero if its length is smaller than 
    // orthogonalization_tolerance()*model.scale(i)
    double orthogonalization_tolerance() const
    { return orthogonalization_tolerance_; }
    void orthogonalization_tolerance(double t)
    { orthogonalization_tolerance_ = t; }


    // get log likelihood function object
    void get_log_likelihood_function(LogLikelihoodFunction& f,
                                     Model& model,
                                     const ObservableVector& cvals)
    { f = LogLikelihoodFunction(*this, model, cvals, minimizer_->verbosity()); }
    void get_log_likelihood_function(LogLikelihoodFunction& f,
                                     Model& model) {
        get_log_likelihood_function
            (f, model, input_function().central_values());
    }
                                     

    // get/set the minimizer
    const BasicMinimizer& minimizer() const { return *minimizer_; }
    BasicMinimizer& minimizer() { return *minimizer_; }
    void minimizer(const BasicMinimizer& mini) {
        if(minimizer_) delete minimizer_;
        minimizer_ = mini.copy();
    }

    // get/set verbosity level of the gsl minimizer
    int minimizer_verbosity() const { return minimizer_verbose_; }
    void minimizer_verbosity(int n) { minimizer_verbose_ = n; }

    // get/set the initial step width of the gsl minimizer
    double minimizer_step_width() const { return minimizer_step_width_; }
    void minimizer_step_width(double s) { minimizer_step_width_ = s; }

    // get/set the tolerance of the gsl minimizer for 1d minimizations.
    // According to gsl documentation a value of 0.1 is usually sufficient
    double minimizer_line_search_precision() const
    { return minimizer_line_search_precision_; }
    void minimizer_line_search_precision(double t)
    { minimizer_line_search_precision_ = t; }

    // get/set the precision with which the minimizer determines the minimum.
    // The exact meaning of the parameter depends on the minimizer.
    // See the gsl documentation for more information.
    double minimizer_precision() const { return minimizer_precision_; }
    void minimizer_precision(double p) { minimizer_precision_ = p; }

    // get/set the maximum number of iterations for function minimization
    int minimizer_iterations() const { return minimizer_iterations_; }
    void minimizer_iterations(int i) { minimizer_iterations_ = i; }

    // if false, the (estimate of) the inverse hessian is set to identity
    // at the start of each minimization.
    bool minimizer_keep_hessian() const { return minimizer_keep_hessian_; }
    void minimizer_keep_hessian(bool b) { minimizer_keep_hessian_ = b; }

    // get/set the initial fraction of points for sampling the inner region
    // in a p-value integral
    double inner_region_fraction() const { return inner_region_fraction_; }
    void inner_region_fraction(double irf) { inner_region_fraction_ = irf; }

    // get/set the power used to construct a mapping for the inner region
    // in a p-value integral
    double inner_region_power() const { return inner_region_power_; }
    void inner_region_power(double irp) { inner_region_power_ = irp; }

    // get/set the chisquare offset for adaptive scans
    double scans_chisquare_offset() const { return scans_chisquare_offset_; }
    void scans_chisquare_offset(double d) { scans_chisquare_offset_ = d; }

    // get/set the (inverse) chisquare power for adaptive scans
    double scans_chisquare_power() const { return scans_chisquare_power_; }
    void scans_chisquare_power(double d) { scans_chisquare_power_ = d; }

    // get/set the number of bins for adaptive scans
    int scans_nbins() const { return scans_nbins_; }
    void scans_nbins(int n) { scans_nbins_ = n; }

    // get/set the verbosity level for p-value integrations
    int dvegas_verbosity() const { return dvegas_verbose_; }
    void dvegas_verbosity(int n) { dvegas_verbose_ = n; }

    // get/set the number of bins for p-value integrations
    int dvegas_nbins() const { return dvegas_nbins_; }
    void dvegas_nbins(int n) { dvegas_nbins_ = n; }

    // get/set the initial number of shots for p-value integrations
    int dvegas_nfirstshots() const { return dvegas_nfirstshots_; }
    void dvegas_nfirstshots(int n) { dvegas_nfirstshots_ = n; }

    // get/set the number of shots per iteration for p-value integrations
    int dvegas_nshots() const { return dvegas_nshots_; }
    void dvegas_nshots(int n) { dvegas_nshots_ = n; }

    // get/set the maximum number iterations for p-value integrations
    int dvegas_niterations() const { return dvegas_niterations_; }
    void dvegas_niterations(int n) { dvegas_niterations_ = n; }

    // get/set the desired relative precision for p-value integrations
    double dvegas_precision() const { return dvegas_precision_; }
    void dvegas_precision(double p) { dvegas_precision_ = p; }

    // toggle adaptation for p-value integrations
    bool dvegas_adaptive() const { return dvegas_adaptive_; }
    void dvegas_adaptive(bool b) { dvegas_adaptive_ = b; }

    // calculate chi-square value using predictions of model with current
    // parameters and central observable values cvals. Stores the result in
    // model's chisquare() member.
    int calc(Model& model, const ObservableVector& cvals) {
        int status;
        if((status = model.calc())) return status;
        model.chisquare(ifunc_.calc(model.observables(),cvals,false)
                        -ifunc_.calc(cvals,cvals,false));
        model.constraint_penalty
            (ifunc_.calc_constraint_penalty(model.observables(), cvals)
             -ifunc_.calc_constraint_penalty(cvals,cvals));
        return 0;
    }

    // calculate chi-square value using predictions of model with current
    // parameters and the default central values from the likelihood
    // function. Stores the result in model's chisquare() member.
    int calc(Model& model)
    { return calc(model, input_function().central_values()); }

    // calculate only chi-square contributions of certain observables
    int calc_contrib(const IndexVector& iv, Model& model,
                     const ObservableVector& cvals)
    {
        int status;
        if((status = model.calc())) return status;
        model.chisquare(ifunc_.calc_contrib(iv,model.observables(),cvals,false)
                        -ifunc_.calc_contrib(iv,cvals,cvals,false));
        model.constraint_penalty
            (ifunc_.calc_penalty_contrib(iv, model.observables(), cvals)
             -ifunc_.calc_penalty_contrib(iv,cvals,cvals));
        return 0;
    }
    int calc_contrib(const IndexVector& iv, Model& model)
    { return calc_contrib(iv, model, input_function().central_values()); }
    int calc_contrib(int iobs, Model& model, const ObservableVector& cvals)
    { return calc_contrib(IndexVector(1, iobs), model, cvals); }
    int calc_contrib(int iobs, Model& model)
    { return calc_contrib(iobs, model, input_function().central_values()); }
    int calc_contrib(int i1, int i2,
                     Model& model, const ObservableVector& cvals)
    {
        assert(i1 >= 0);
        assert(i2 >= i1);
        IndexVector iv(i2-i1+1);
        for(int i=i1; i <= i2; ++i) iv[i-i1] = i;
        return calc_contrib(iv, model, cvals);
    }
    int calc_contrib(int i1, int i2, Model& model)
    { return calc_contrib(i1, i2, model, input_function().central_values()); }

    // minimize the likelihood-function with respect to the floating parameters.
    // The data from the last scan() is used to find the best start parameters.
    // Information about the best-fit point is stored in model().
    virtual int global_fit(Model& model, const ObservableVector& cvals)
    { get_starting_point(model, cvals); return local_fit(model, cvals); }
    int global_fit(Model& model)
    { return global_fit(model, input_function().central_values()); }

    // minimize the likelihood-function with respect to the floating parameters.
    // The current parameters of model() are used as start parameters for the
    // minimization. Information about the best-fit point is stored in model().
    virtual int local_fit(Model& model, const ObservableVector& cvals);
    int local_fit(Model& model)
    { return local_fit(model, input_function().central_values()); }

    // searches through the model's dictionary (built up from calls to scan())
    // to find the best starting point for a minimization. The parameters of
    // the model object are set to that starting point.
    virtual void get_starting_point(Model& model,
                                    const ObservableVector& cvals);
    void get_starting_point(Model& model)
    { return get_starting_point(model, input_function().central_values()); }

    // returns a ScanIntegrand object for adaptively scanning model model
    // for central input values cvals. 
    ScanIntegrand get_scan_integrand(Model& model,
                                     const ObservableVector& cvals) {
        return ScanIntegrand(model, ifunc_, cvals,
                             scans_chisquare_offset_, scans_chisquare_power_);
    }
    ScanIntegrand get_scan_integrand(Model& model)
    {  return get_scan_integrand(model, input_function().central_values()); }

    // adaptively scans model model with an initial number of nfirstshots
    // shots and a subsequent number of niter iterations with nshots shots
    // each. The dictionary of model is then filled with a final iteration with
    // nlastshots shots.
    void adaptive_scan(Model& model, const ObservableVector& cvals,
                       HepSource::Int64 nfirstshots, HepSource::Int64 nshots,
                       int niter, HepSource::Int64 nlastshots);
    void adaptive_scan(Model& model,
                       HepSource::Int64 nfirstshots, HepSource::Int64 nshots,
                       int niter, HepSource::Int64 nlastshots) {
        adaptive_scan(model, input_function().central_values(),
                      nfirstshots, nshots, niter, nlastshots);
    }

    // compute the statistical significance (p-value) of a likelihood ratio
    // test for nested models
    double calc_nested_lrt_pvalue(Model& fullmod, Model& constrainedmod);

    // compute the statistical significance (p-value) of a likelihood ratio
    // test for non-nested models. 
    double calc_lrt_pvalue(Model& mod0, Model& mod1, int argc=0, char** argv=0);

    // returns the result of the last p-value integration
    double pvalue() const { return pvalue_; }

    // returns the error of the last p-value integration
    double pvalue_error() const { return pvalue_error_; }

    // checks if the precision goal was reached in the last p-value integration
    bool reached_precision_goal() const { return reached_precision_goal_; }

    // return total number of shots from last p-value integration
    HepSource::Int64 number_of_shots() const { return nshots_; }

    // return fraction of failed shots in last p-value integration
    double failed_shot_ratio() const { return failed_shot_ratio_; }
};


}

#endif
