#ifndef MYFITTER_GRIDS_GRIDINTERPOLATOR_HPP
#define MYFITTER_GRIDS_GRIDINTERPOLATOR_HPP

#include <iostream>
#include <limits>
#include <map>
#include <algorithm>
#include <sstream>
#include <cmath>
#include <fstream>

namespace myfitter { namespace grids {

    template<class TX, class TY>
    class IrregularGridInterpolator1D {
    private:
        bool extrapolate_;
        std::map<TX, TY> map_;

        TY lin_extrapolate_
        (TX x, typename std::map<TX, TY>::const_iterator upper) const {
            if(!extrapolate_) return std::numeric_limits<TY>::quiet_NaN();
            if(map_.size() == 1) return map_.begin()->second;

            typename std::map<TX, TY>::const_iterator lower = upper;
            if(upper==map_.begin())
                upper++;
            else if(upper==map_.end()) {
                upper--;
                lower--; lower--;
            } else
                return std::numeric_limits<TY>::quiet_NaN();

            return lower->second + (upper->second-lower->second)/
                (upper->first-lower->first)*(x-lower->first);
        }

        TY spline_extrapolate_
        (TX x, typename std::map<TX, TY>::const_iterator upper) const {
            if(!extrapolate_) return std::numeric_limits<TY>::quiet_NaN();
            if(map_.size() < 2) return lin_extrapolate_(x, upper);

            typename std::map<TX, TY>::const_iterator lower, lower2, upper2;
            if(upper==map_.begin()) {
                lower = upper;
                upper2 = ++upper;
                ++upper2;

                TY v0 = (upper2->second - lower->second)/
                    (upper2->first - lower->first);
                TX x1 = lower->first - upper->first;
                TY y0 = upper->second;
                TX dx = x - upper->first;
                TY y1 = lower->second;

                TY a = 2.*(y1-y0 - v0*x1)/x1 + v0;
                return y1 + a*(dx-x1);
            } else if(upper==map_.end()) {
                lower = --upper;
                lower2 = --lower;
                --lower2;

                TY v0 = (upper->second - lower2->second)/
                    (upper->first - lower2->first);
                TX x1 = upper->first - lower->first;
                TY y0 = lower->second;
                TX dx = x-lower->first;
                TY y1 = upper->second;

                TY a = 2.*(y1-y0 - v0*x1)/x1 + v0;
                return y1 + a*(dx-x1);
            } else
                return std::numeric_limits<TY>::quiet_NaN();

        }

    public:
        // constructors
        IrregularGridInterpolator1D() : extrapolate_(false) { }
        IrregularGridInterpolator1D(const IrregularGridInterpolator1D& g) :
            extrapolate_(g.extrapolate_), map_(g.map_) { }

        // assignment operator
        virtual IrregularGridInterpolator1D& operator()
            (const IrregularGridInterpolator1D& g) {
            extrapolate_ = g.extrapolate_;
            map_ = g.map_;
            return *this;
        }

        // add points
        void set(TX x, TY y) { map_[x] = y; }

        // clear data
        void clear() { map_.clear(); }

        // toggle extrapolation
        bool extrapolate() const { return extrapolate_; }
        void extrapolate(bool b) { extrapolate_ = b; }

        // read data
        bool read(std::istream& is, int col=0) {
            std::string line, dump;
            std::istringstream ls;
            ls.setf(std::ios::skipws);
            std::string::size_type first;

            while(std::getline(is, line)) {
                ls.str(line); ls.clear();

                if((first = line.find_first_not_of(" \t\r\n"))
                   == std::string::npos || line[first] == '#')
                    continue;

                TX x;
                TY y;

                ls >> x;
                for(int j=0; j<col; ++j)
                    ls >> dump;
                ls >> y;
                if(!ls) return false;

                set(x, y);
            }

            return true;
        }

        bool read(std::string s, int col=0) {
            std::ifstream is;
            is.open(s.c_str());
            bool success = read(is,col);
            is.close();
            return success;
        }

        // get minimum and maximum x value

        TX min() const {
            if(map_.empty()) return std::numeric_limits<TY>::quiet_NaN();
            return map_.begin()->first;
        }
        TX max() const {
            if(map_.empty()) return std::numeric_limits<TY>::quiet_NaN();
            return map_.rbegin()->first;
        }

        // check if x value is contained in range

        bool contains(TX x) const { return x >= min() && x <= max(); }


        // linear interpolation

        TY lin_interp(TX x) const {
            if(map_.empty()) return std::numeric_limits<TY>::quiet_NaN();

            typename std::map<TX, TY>::const_iterator lower, upper;
            upper = map_.upper_bound(x);

            if(upper==map_.begin())
                return lin_extrapolate_(x, upper);

            if(upper==map_.end()) {
                if(map_.rbegin()->first == x)
                    return map_.rbegin()->second;
                else
                    return lin_extrapolate_(x, upper);
            }
            
            lower = upper;
            --lower;

            TX r = x/(upper->first - lower->first);
            return (1.-r)*lower->second + r*upper->second;
        }

        TY spline_interp(TX x) const {
            if(map_.empty()) return std::numeric_limits<TY>::quiet_NaN();

            typename std::map<TX, TY>::const_iterator lower, lower2,
                upper, upper2;
            upper = map_.upper_bound(x);

            if(upper==map_.begin())
                return spline_extrapolate_(x, upper);

            if(upper==map_.end()) {
                if(map_.rbegin()->first == x)
                    return map_.rbegin()->second;
                else
                    return spline_extrapolate_(x, upper);
            }
            
            lower = upper;
            --lower;

            if(lower==map_.begin() && upper==(--map_.end())) {
                TX r = x/(upper->first - lower->first);
                return (1.-r)*lower->second + r*upper->second;
            }

            if(lower==map_.begin()) {
                upper2 = upper; ++upper2;
                TY v0 = (upper2->second - lower->second)/
                    (upper2->first - lower->first);
                TX x1 = lower->first - upper->first;
                TY y0 = upper->second;
                TX dx = x - upper->first;
                TY y1 = lower->second;
                return (y1-y0 - v0*x1)*dx*dx/(x1*x1) + v0*dx + y0;
            }

            if(upper==(--map_.end())) {
                lower2 = lower; --lower2;
                TY v0 = (upper->second - lower2->second)/
                    (upper->first - lower2->first);
                TX x1 = upper->first - lower->first;
                TY y0 = lower->second;
                TX dx = x-lower->first;
                TY y1 = upper->second;
                return (y1-y0 - v0*x1)*dx*dx/(x1*x1) + v0*dx + y0;
            }

            lower2 = lower; --lower2;
            upper2 = upper; ++upper2;
            TY v0 = (upper->second - lower2->second)/
                (upper->first - lower2->first);
            TY v1 = (upper2->second - lower->second)/
                (upper2->first - lower->first);
            TX x10 = upper->first - lower->first;
            TY y0 = lower->second;
            TY y1 = upper->second;
            TY y10 = y1-y0;
            TX xx0 = x - lower->first;
            TX x1x = upper->first - x;

            TY a = -(v1*x10-y10);
            TY b = (v0*x10-y10);

            return (y0*x1x + y1*xx0)/x10
                + xx0*x1x*(a*xx0 + b*x1x)/(x10*x10*x10);
        }
    };

    template<int N, class TX, class TY>
    class BasicGridInterpolator {
    private:
        TY* data_;
        int n_[N];
        TX d_[N];
        TX b_[N];

        TY interp_(TX r[], int i[], int n) const {
            if(n == 1) {
                if(i[N-n] >= n_[N-n] - 1) return data_[idata_(i)];
                TY v0 = data_[idata_(i)];
                i[N-n] += 1;
                TY v1 = data_[idata_(i)];
                i[N-n] -= 1;
                return (1.-r[0])*v0 + r[0]*v1;
            } else {
                if(i[N-n] >= n_[N-n] - 1) return data_[idata_(i)];
                TY v0 = interp_(r+1, i, n-1);
                i[N-n] += 1;
                TY v1 = interp_(r+1, i, n-1);
                i[N-n] -= 1;
                return (1.-r[0])*v0 + r[0]*v1;
            }
        }

        TY interp3_(TX r[], int i[], int n) const {
            if(n == 0)
                return data_[idata_(i)];

            int j = i[N-n];
            if(j == 0) {
                if(j >= n_[N-n] - 2) {
                    TY v0 = interp3_(r+1, i, n-1);
                    i[N-n] += 1;
                    TY v1 = interp3_(r+1, i, n-1);
                    i[N-n] -= 1;
                    return (1.-r[0])*v0 + r[0]*v1;
                } else {
                    TY v0 = interp3_(r+1, i, n-1);
                    i[N-n] += 1;
                    TY v1 = interp3_(r+1, i, n-1);
                    i[N-n] += 1;
                    TY k1 = (interp3_(r+1, i, n-1) - v0)*0.5;
                    i[N-n] -= 2;
                    TX dr = r[0] - 1;
                    return (v0 - v1 + k1)*dr*dr + k1*dr + v1;
                }
            } else if (j >= n_[N-n] - 2) {
                if(j >= n_[N-n] - 1) return data_[idata_(i)];
                TY v0 = interp3_(r+1, i, n-1);
                i[N-n] += 1;
                TY v1 = interp3_(r+1, i, n-1);
                i[N-n] -= 2;
                TY k0 = (v1 - interp3_(r+1, i, n-1))*0.5;
                i[N-n] += 1;
                return (v1 - v0 - k0)*r[0]*r[0] + k0*r[0] + v0;
            } else {
                TY v0 = interp3_(r+1, i, n-1);
                i[N-n] += 1;
                TY v1 = interp3_(r+1, i, n-1);
                i[N-n] += 1;
                TY k1 = (interp3_(r+1, i, n-1) - v0)*0.5;
                i[N-n] -= 3;
                TY k0 = (v1 - interp3_(r+1, i, n-1))*0.5;
                i[N-n] += 1;

                TY a = k0 - (v1 - v0);
                TY b = -k1 + (v1 - v0);
                return (1.-r[0])*v0 + r[0]*v1 +
                    r[0]*(1.-r[0])*(a*(1.-r[0]) + b*r[0]);
            }
        }

    protected:
        int idata_(int i[N]) const {
            int k=i[0];
            int nn = 1;
            for(int j=1; j<N; ++j) {
                nn *= n_[j-1];
                k += nn*i[j];
            }
            return k;
        }

        void geti_(int k, int i[N]) const {
            int nn[N];
            nn[0] = 1;
            for(int j=1; j<N; ++j)
                nn[j] = nn[j-1]*n_[j-1];
            for(int j=N-1; j>=0; --j) {
                i[j] = k/nn[j];
                k = k-i[j]*nn[j];
            }
        }

        void resize_() {
            int nn=1;
            for(int j=0; j<N; ++j)
                nn *= n_[j];
            if(data_ != 0) delete[] data_;
            data_ = new TY[nn];
            for(int j=0; j<nn; ++j)
                data_[j] = std::numeric_limits<TY>::quiet_NaN();
        }
        

    public:
        // constructors
        BasicGridInterpolator() : data_(0)
        { for(int i=0; i<N; ++i) { n_[i] = 0; b_[i] = 0.; d_[i] = 0.; } }
        BasicGridInterpolator(const BasicGridInterpolator<N, TX, TY>& g) {
            int nn=1;
            for(int j=0; j<N; ++j) {
                b_[j] = g.b_[j];
                d_[j] = g.d_[j];
                n_[j] = g.n_[j];
                nn *= n_[j];
            }
            if(g.data_ == 0) { data_ = 0; return; }
            data_ = new TY[nn];
            for(int j=0; j<nn; ++j)
                data_[j] = g.data_[j];
        }
        virtual const BasicGridInterpolator<N, TX, TY>&
        operator=(const BasicGridInterpolator<N, TX, TY>& g) {
            free();
            int nn=1;
            for(int j=0; j<N; ++j) {
                b_[j] = g.b_[j];
                d_[j] = g.d_[j];
                n_[j] = g.n_[j];
                nn *= n_[j];
            }
            if(g.data_ == 0) { data_ = 0; return *this; }
            data_ = new TY[nn];
            for(int j=0; j<nn; ++j)
                data_[j] = g.data_[j];

            return *this;
        }

        // destructor
        ~BasicGridInterpolator() { free(); }

        // setup grid
        void resize(TX base[N], TX step[N], int n[N]) {
            for(int j=0; j<N; ++j) {
                b_[j] = base[j];
                d_[j] = step[j];
                n_[j] = n[j];
            }
            resize_();
        }

        // free memory
        void free() {
            for(int i=0; i<N; ++i) { n_[i] = 0; b_[i] = 0.; d_[i] = 0.; }
            if(data_ != 0) delete[] data_;
            data_ = 0;
        }

        // grid information
        TX base(int i) const {
            if(i<0 || i >= N)
                return std::numeric_limits<TX>::quiet_NaN();
            else
                return b_[i];
        }
        TX step(int i) const {
            if(i<0 || i >= N)
                return std::numeric_limits<TX>::quiet_NaN();
            else
                return d_[i];
        }
        int size(int i) const {
            if(i<0 || i >= N)
                return -1;
            else
                return n_[i];
        }
        TX min(int i) const { return base(i); }
        TX max(int i) const { return b_[i]+(n_[i]-1)*d_[i]; }
        


        // read data
        bool read(std::istream& is, int col=0, double prec=0.001) {
            std::string line, dump;
            std::istringstream ls;
            ls.setf(std::ios::skipws);
            std::string::size_type first;

            do
                if(!std::getline(is, line)) return false;
            while((first = line.find_first_not_of(" \t\r\n"))
                  == std::string::npos || line[first] == '#');
            ls.str(line); ls.clear();
            for(int i=0; i<N; ++i)
                ls >> b_[i];
            if(!ls) return false;

            do
                if(!std::getline(is, line)) return false;
            while((first = line.find_first_not_of(" \t\r\n"))
                  == std::string::npos || line[first] == '#');
            ls.str(line); ls.clear();
            for(int i=0; i<N; ++i)
                ls >> d_[i];
            if(!ls) return false;

            do
                if(!std::getline(is, line)) return false;
            while((first = line.find_first_not_of(" \t\r\n"))
                  == std::string::npos || line[first] == '#');
            ls.str(line); ls.clear();
            for(int i=0; i<N; ++i)
                ls >> n_[i];
            if(!ls) return false;

            resize_();

            while(std::getline(is, line)) {
                ls.str(line); ls.clear();

                if((first = line.find_first_not_of(" \t\r\n"))
                   == std::string::npos || line[first] == '#')
                    continue;

                TX x[N];
                TX r[N];
                int i[N];
                TY val;
                bool skip = false;
                for(int j=0; j<N; ++j) {
                    ls >> x[j];
                    r[j] = (x[j]-b_[j])/d_[j];
                    i[j] = r[j]+0.5;
                    if(std::abs(i[j]-r[j]) > prec) { skip=true; break; }
                    if(i[j] >= n_[j]) { skip=true; break; }
                }
                if(skip) continue;
                for(int j=0; j<col; ++j)
                    ls >> dump;
                ls >> val;
                if(!ls) return false;

                data_[idata_(i)] = val;
            }

            return true;
        }

        bool read(std::string s, int col=0, double prec=0.001) {
            std::ifstream is;
            is.open(s.c_str());
            bool success = read(is,col,prec);
            is.close();
            return success;
        }

        bool check() const {
            int n=1;
            bool result=true;
            for(int j=0; j<N; ++j) n *= n_[j];
            for(int j=0; j<n; ++j)
                if(isnan(data_[j])) {
                    int i[N];
                    std::cerr << "BasicGridInterpolator::check(): NaN encountered at ";
                    geti_(j, i);
                    for(int k=0; k<N; ++k)
                        std::cerr << b_[k] + i[k]*d_[k] << " ";
                    std::cerr << std::endl;
                    result = false;
                }
            return result;
        }

        bool contains(TX x[N]) const {
            for(int j=0; j<N; ++j)
                if(x[j] < b_[j] || x[j] > b_[j]+(n_[j]-1)*d_[j]) return false;
            return true;
        }

        TY lin_interp(TX x[N]) const {
            if(!contains(x)) return std::numeric_limits<TY>::quiet_NaN();
    
            TX r[N];
            for(int j=0; j<N; ++j)
                r[j] = (x[j]-b_[j])/d_[j];

            // i: grid point coordinates
            // r: coordinates in grid cube (between 0 and 1)
            int i[N];
            for(int j=0; j<N; ++j) {
                i[j] = r[j];
                r[j] -= i[j];
            }

            // p: permutation that brings the r[j] in descending order
            int p[N];
            for(int j=0; j<N; ++j) p[j] = j;
            for(int j=0; j<N-1; ++j)
                for(int k=j+1; k<N; ++k)
                    if(x[p[k]] > x[p[j]]) std::swap(p[k], p[j]);

            // val: function values at corners of grid cube
            TY val[N+1];
            val[0] = data_[idata_(i)];
            for(int j=0; j<N; ++j) {
                i[p[j]] += 1;
                val[j+1] = data_[idata_(i)];
            }

            for(int j=0; j<N; ++j)
                val[0] += (val[j+1]-val[j])*r[p[j]];

            return val[0];
        }

        TY interp(TX x[N]) const {
            if(!contains(x)) return std::numeric_limits<TY>::quiet_NaN();

            TX r[N];
            for(int j=0; j<N; ++j)
                r[j] = (x[j]-b_[j])/d_[j];

            // i: grid point coordinates
            // r: coordinates in grid cube (between 0 and 1)
            int i[N];
            for(int j=0; j<N; ++j) {
                i[j] = r[j];
                r[j] -= i[j];
            }
            
            return interp_(r,i,N);
        }

        TY spline_interp(TX x[N]) const {
            if(!contains(x)) return std::numeric_limits<TY>::quiet_NaN();

            TX r[N];
            for(int j=0; j<N; ++j)
                r[j] = (x[j]-b_[j])/d_[j];

            // i: grid point coordinates
            // r: coordinates in grid cube (between 0 and 1)
            int i[N];
            for(int j=0; j<N; ++j) {
                i[j] = r[j];
                r[j] -= i[j];
            }
            
            return interp3_(r,i,N);
        }

        TY get(int i[N]) const {
            if(data_ == 0) return std::numeric_limits<TY>::quiet_NaN();
            for(int j=0; j<N; ++j)
                if(i[j] < 0 || i[j] >= n_[j])
                    return std::numeric_limits<TY>::quiet_NaN();
            return data_[idata_(i)];
        }

        TY set(int i[N], TY val) {
            if(data_ == 0) return std::numeric_limits<TY>::quiet_NaN();
            for(int j=0; j<N; ++j)
                if(i[j] < 0 || i[j] >= n_[j])
                    return std::numeric_limits<TY>::quiet_NaN();
            return data_[idata_(i)] = val;
        }
    };


    template<class TX, class TY>
    class GridInterpolator1D : public BasicGridInterpolator<1,TX,TY> {
    public:
        // constructors
        GridInterpolator1D() : BasicGridInterpolator<1,TX,TY>() { }
        GridInterpolator1D(TX base1, TX step1, int n1) :
            BasicGridInterpolator<1,TX,TY>()
        { resize(base1, step1, n1); }
        GridInterpolator1D(const GridInterpolator1D<TX, TY>& g) :
            BasicGridInterpolator<1,TX,TY>(g) { }

        // resize function
        void resize(TX base1, TX step1, int n1) {
            TX base[1];
            TX step[1];
            int n[1];
            base[0] = base1;
            step[0] = step1;
            n[0] = n1;
            BasicGridInterpolator<1,TX,TY>::resize(base, step, n);
        }

        TY lin_interp(TX x1) const {
            TX x[1];
            x[0] = x1;
            return BasicGridInterpolator<1,TX,TY>::lin_interp(x);
        }

        TY interp(TX x1) const {
            TX x[1];
            x[0] = x1;
            return BasicGridInterpolator<1,TX,TY>::interp(x);
        }

        TY spline_interp(TX x1) const {
            TX x[1];
            x[0] = x1;
            return BasicGridInterpolator<1,TX,TY>::spline_interp(x);
        }

        bool contains(TX x1) const {
            TX x[1];
            x[0] = x1;
            return BasicGridInterpolator<1,TX,TY>::contains(x);
        }

        TY get(int i1) const {
            int i[1];
            i[0] = i1;
            return BasicGridInterpolator<1,TX,TY>::get(i);
        }

        TY set(int i1, TY val) {
            int i[1];
            i[0] = i1;
            return BasicGridInterpolator<1,TX,TY>::set(i,val);
        }
    };

    template<class TX, class TY>
    class GridInterpolator2D : public BasicGridInterpolator<2,TX,TY> {
    public:
        // constructors
        GridInterpolator2D() : BasicGridInterpolator<2,TX,TY>() { }
        GridInterpolator2D(TX base1, TX step1, int n1,
                           TX base2, TX step2, int n2) :
            BasicGridInterpolator<2,TX,TY>()
        { resize(base1, step1, n1, base2, step2, n2); }
        GridInterpolator2D(const GridInterpolator2D<TX, TY>& g) :
            BasicGridInterpolator<2,TX,TY>(g) { }

        // resize function
        void resize(TX base1, TX step1, int n1,
                    TX base2, TX step2, int n2) {
            TX base[2];
            TX step[2];
            int n[2];
            base[0] = base1; base[1] = base2;
            step[0] = step1; step[1] = step2;
            n[0] = n1; n[1] = n2;
            BasicGridInterpolator<2,TX,TY>::resize(base, step, n);
        }

        TY lin_interp(TX x1, TX x2) const {
            TX x[2];
            x[0] = x1; x[1] = x2;
            return BasicGridInterpolator<2,TX,TY>::lin_interp(x);
        }

        TY interp(TX x1, TX x2) const {
            TX x[2];
            x[0] = x1; x[1] = x2;
            return BasicGridInterpolator<2,TX,TY>::interp(x);
        }

        TY spline_interp(TX x1, TX x2) const {
            TX x[2];
            x[0] = x1; x[1] = x2;
            return BasicGridInterpolator<2,TX,TY>::spline_interp(x);
        }

        bool contains(TX x1, TX x2) const {
            TX x[2];
            x[0] = x1; x[1] = x2;
            return BasicGridInterpolator<2,TX,TY>::contains(x);
        }

        TY get(int i1, int i2) const {
            int i[2];
            i[0] = i1; i[1] = i2;
            return BasicGridInterpolator<2,TX,TY>::get(i);
        }

        TY set(int i1, int i2, TY val) {
            int i[2];
            i[0] = i1; i[1] = i2;
            return BasicGridInterpolator<2,TX,TY>::set(i,val);
        }
    };

    template<class TX, class TY>
    class GridInterpolator3D : public BasicGridInterpolator<3,TX,TY> {
    public:
        // constructors
        GridInterpolator3D() : BasicGridInterpolator<3,TX,TY>() { }
        GridInterpolator3D(TX base1, TX step1, int n1,
                           TX base2, TX step2, int n2,
                           TX base3, TX step3, int n3) :
            BasicGridInterpolator<3,TX,TY>()
        { resize(base1, step1, n1, base2, step2, n2, base3, step3, n3); }
        GridInterpolator3D(const GridInterpolator3D<TX, TY>& g) :
            BasicGridInterpolator<3,TX,TY>(g) { }

        // resize function
        void resize(TX base1, TX step1, int n1,
                    TX base2, TX step2, int n2,
                    TX base3, TX step3, int n3) {
            TX base[3];
            TX step[3];
            int n[3];
            base[0] = base1; base[1] = base2; base[2] = base3;
            step[0] = step1; step[1] = step2; step[2] = step3;
            n[0] = n1; n[1] = n2; n[2] = n3;
            BasicGridInterpolator<3,TX,TY>::resize(base, step, n);
        }

        TY lin_interp(TX x1, TX x2, TX x3) const {
            TX x[3];
            x[0] = x1; x[1] = x2; x[2] = x3;
            return BasicGridInterpolator<3,TX,TY>::lin_interp(x);
        }

        TY interp(TX x1, TX x2, TX x3) const {
            TX x[3];
            x[0] = x1; x[1] = x2; x[2] = x3;
            return BasicGridInterpolator<3,TX,TY>::interp(x);
        }

        TY spline_interp(TX x1, TX x2, TX x3) const {
            TX x[3];
            x[0] = x1; x[1] = x2; x[2] = x3;
            return BasicGridInterpolator<3,TX,TY>::spline_interp(x);
        }

        bool contains(TX x1, TX x2, TX x3) const {
            TX x[3];
            x[0] = x1; x[1] = x2; x[2] = x3;
            return BasicGridInterpolator<3,TX,TY>::contains(x);
        }

        TY get(int i1, int i2, int i3) const {
            int i[3];
            i[0] = i1; i[1] = i2; i[2] = i3;
            return BasicGridInterpolator<3,TX,TY>::get(i);
        }

        TY set(int i1, int i2, int i3, TY val) {
            int i[3];
            i[0] = i1; i[1] = i2; i[2] = i3;
            return BasicGridInterpolator<3,TX,TY>::set(i,val);
        }
    };

    template<class TX, class TY>
    class GridInterpolator4D : public BasicGridInterpolator<4,TX,TY> {
    public:
        // constructors
        GridInterpolator4D() : BasicGridInterpolator<4,TX,TY>() { }
        GridInterpolator4D(TX base1, TX step1, int n1,
                           TX base2, TX step2, int n2,
                           TX base3, TX step3, int n3,
                           TX base4, TX step4, int n4) :
            BasicGridInterpolator<4,TX,TY>()
        { resize(base1, step1, n1, base2, step2, n2,
                 base3, step3, n3, base4, step4, n4); }
        GridInterpolator4D(const GridInterpolator4D<TX, TY>& g) :
            BasicGridInterpolator<4,TX,TY>(g) { }

        // resize function
        void resize(TX base1, TX step1, int n1,
                    TX base2, TX step2, int n2,
                    TX base3, TX step3, int n3,
                    TX base4, TX step4, int n4) {
            TX base[4];
            TX step[4];
            int n[4];
            base[0] = base1; base[1] = base2; base[2] = base3; base[3] = base4;
            step[0] = step1; step[1] = step2; step[2] = step3; step[3] = step4;
            n[0] = n1; n[1] = n2; n[2] = n3; n[3] = n4;
            BasicGridInterpolator<4,TX,TY>::resize(base, step, n);
        }

        TY lin_interp(TX x1, TX x2, TX x3, TX x4) const {
            TX x[4];
            x[0] = x1; x[1] = x2; x[2] = x3; x[3] = x4;
            return BasicGridInterpolator<4,TX,TY>::lin_interp(x);
        }

        TY interp(TX x1, TX x2, TX x3, TX x4) const {
            TX x[4];
            x[0] = x1; x[1] = x2; x[2] = x3; x[3] = x4;
            return BasicGridInterpolator<4,TX,TY>::interp(x);
        }

        TY spline_interp(TX x1, TX x2, TX x3, TX x4) const {
            TX x[4];
            x[0] = x1; x[1] = x2; x[2] = x3; x[3] = x4;
            return BasicGridInterpolator<4,TX,TY>::spline_interp(x);
        }

        bool contains(TX x1, TX x2, TX x3, TX x4) const {
            TX x[4];
            x[0] = x1; x[1] = x2; x[2] = x3; x[3] = x4;
            return BasicGridInterpolator<4,TX,TY>::contains(x);
        }

        TY get(int i1, int i2, int i3, int i4) const {
            int i[4];
            i[0] = i1; i[1] = i2; i[2] = i3; i[3] = i4;
            return BasicGridInterpolator<4,TX,TY>::get(i);
        }

        TY set(int i1, int i2, int i3, int i4, TY val) {
            int i[4];
            i[0] = i1; i[1] = i2; i[2] = i3; i[3] = i4;
            return BasicGridInterpolator<4,TX,TY>::set(i,val);
        }
    };

    template<class TX, class TY>
    class GridInterpolator5D : public BasicGridInterpolator<5,TX,TY> {
    public:
        // constructors
        GridInterpolator5D() : BasicGridInterpolator<5,TX,TY>() { }
        GridInterpolator5D(TX base1, TX step1, int n1,
                           TX base2, TX step2, int n2,
                           TX base3, TX step3, int n3,
                           TX base4, TX step4, int n4,
                           TX base5, TX step5, int n5) :
            BasicGridInterpolator<5,TX,TY>()
        { resize(base1, step1, n1, base2, step2, n2,
                 base3, step3, n3, base4, step4, n4,
                 base5, step5, n5); }

        // resize function
        void resize(TX base1, TX step1, int n1,
                    TX base2, TX step2, int n2,
                    TX base3, TX step3, int n3,
                    TX base4, TX step4, int n4,
                    TX base5, TX step5, int n5) {
            TX base[5];
            TX step[5];
            int n[5];
            base[0] = base1; base[1] = base2; base[2] = base3; base[3] = base4;
            base[4] = base5;
            step[0] = step1; step[1] = step2; step[2] = step3; step[3] = step4;
            step[4] = step5;
            n[0] = n1; n[1] = n2; n[2] = n3; n[3] = n4; n[4] = n5;
            BasicGridInterpolator<5,TX,TY>::resize(base, step, n);
        }

        TY lin_interp(TX x1, TX x2, TX x3, TX x4, TX x5) const {
            TX x[5];
            x[0] = x1; x[1] = x2; x[2] = x3; x[3] = x4; x[4] = x5;
            return BasicGridInterpolator<5,TX,TY>::lin_interp(x);
        }

        TY interp(TX x1, TX x2, TX x3, TX x4, TX x5) const {
            TX x[5];
            x[0] = x1; x[1] = x2; x[2] = x3; x[3] = x4; x[4] = x5;
            return BasicGridInterpolator<5,TX,TY>::interp(x);
        }

        TY spline_interp(TX x1, TX x2, TX x3, TX x4, TX x5) const {
            TX x[5];
            x[0] = x1; x[1] = x2; x[2] = x3; x[3] = x4; x[4] = x5;
            return BasicGridInterpolator<5,TX,TY>::spline_interp(x);
        }

        bool contains(TX x1, TX x2, TX x3, TX x4, TX x5) const {
            TX x[5];
            x[0] = x1; x[1] = x2; x[2] = x3; x[3] = x4; x[4] = x5;
            return BasicGridInterpolator<5,TX,TY>::contains(x);
        }

        TY get(int i1, int i2, int i3, int i4, int i5) const {
            int i[5];
            i[0] = i1; i[1] = i2; i[2] = i3; i[3] = i4; i[4] = i5;
            return BasicGridInterpolator<5,TX,TY>::get(i);
        }

        TY set(int i1, int i2, int i3, int i4, int i5, TY val) {
            int i[5];
            i[0] = i1; i[1] = i2; i[2] = i3; i[3] = i4; i[4] = i5;
            return BasicGridInterpolator<5,TX,TY>::set(i,val);
        }
    };
}}


#endif
