/*
This file is part of myFitter, a C++ library for maximum likelihood fits.
Copyright (C) 2012, 2013  Martin Wiebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef MYFITTER_INPUTFUNCTION_HPP
#define MYFITTER_INPUTFUNCTION_HPP

#include "model.hpp"


namespace myfitter {

typedef std::vector<int> IndexVector;


// base class for likelihood-components

class InputComponent {
private:
    int nobservables_;
    IndexVector claimed_observables_;
    IndexVector obs_pos_;
    ObservableVector central_values_;

protected:
    // vector of central values. Uses internal indices.

    // claim observable with external index i
    void claim_(int iobs, double val) {
        assert(iobs < nobservables());
        assert(iobs >= 0);
        obs_pos_[iobs] = n_claimed_observables();
        claimed_observables_.push_back(iobs);
        central_values_.resize(claimed_observables_.size());
        central_values_[claimed_observables_.size()-1] = val;
    }

    // access central values with internal indices
    double central_value_(int i) const { return central_values_[i]; }
    void central_value_(int i, double val) { central_values_[i] = val; }

public:
    // constructors
    InputComponent(int nobs) : nobservables_(nobs), obs_pos_(nobs, -1) { }
    InputComponent(const InputComponent& lc) :
        nobservables_(lc.nobservables()),
        claimed_observables_(lc.claimed_observables()),
        obs_pos_(lc.obs_pos_), central_values_(lc.central_values_) { }

    // virtual copy constructor
    // DERIVED CLASSES MUST OVERLOAD THIS!!
    virtual InputComponent* copy() const { return 0; }

    // number of observables required by the likelihood-component.
    // The number of observables in the likelihood-function may be bigger.
    int nobservables() const { return nobservables_; }

    // returns the number of claimed observables
    int n_claimed_observables() const { return claimed_observables_.size(); }

    // access indices of claimed observables.
    // The claimed observables enter this component of the input function.
    const IndexVector& claimed_observables() const
    { return claimed_observables_; }
    int external_index(int i) const {
        assert(i >= 0 && i < n_claimed_observables());
        return claimed_observables_[i];
    }

    // returns internal index corresponding to observable with external index i
    int internal_index(int i) const 
    { assert(obs_pos_[i] >= 0); return obs_pos_[i]; }

    // access vector of central values (with internal indices)
    const ObservableVector& central_values() const { return central_values_; }

    // calculates the likelihood-component.
    // The size of the vector arguments will be n_claimed_observables()
    // Derived classes must implement this.
    virtual double calc(const ObservableVector& pred,
                        const ObservableVector& meas) = 0;

    // calculates the derivatives with respect to the predicted observables
    // (first argument to calc()). Only those entries of result corresponding
    // to claimed observables should be set.
    // The size of the vector arguments will be n_claimed_observables()
    // Derived classes must implement this.
    virtual int calc_deriv(const ObservableVector& pred,
                           const ObservableVector& meas,
                           ObservableVector& result) = 0;

    // calculates the likelihood-component without systematic errors.
    // Likelihood components which implement systematics should overload this
    // function.
    // The size of the vector arguments will be n_claimed_observables()
    virtual double calc_nosystematics(const ObservableVector& pred,
                                      const ObservableVector& meas)
    { return calc(pred, meas); }

    // fill the entries of the matrix hess corresponding to (pairs of) claimed
    // observables with the elements of the Hessian of the likelihood function
    // at pred=meas=obs. The Hessian is the quadratic term in the Taylor
    // expansion of calc_nosystematics(obs, obs), considered as a function of
    // obs (without the factor 1/2).
    // The size of obs and dimensions of hess will be n_claimed_observables()
    virtual void get_hessian(Matrix& hess,
                             const ObservableVector& obs) = 0;
};


// class representing a likelihood function.
// Serves as container for likelihood-components.

class InputFunction {
private:
    int nobs_;
    typedef std::vector<InputComponent*> InputComponentVector;
    InputComponentVector lcv_;
    IndexVector claimed_observables_;
    std::vector<bool> is_constraint_;
    double constraint_penalty_factor_;
    ObservableVector central_values_;

public:
    // constructor
    InputFunction(int nobs) : nobs_(nobs), constraint_penalty_factor_(1.),
                              central_values_(nobs, Model::nan()) { }

    // copy constructor
    InputFunction(const InputFunction& f) :
        nobs_(f.nobs_), lcv_(f.lcv_.size()),
        claimed_observables_(f.claimed_observables_),
        is_constraint_(f.is_constraint_),
        constraint_penalty_factor_(f.constraint_penalty_factor_),
        central_values_(f.central_values_)
    {
        for(int i=0; i < int(f.lcv_.size()); ++i)
            lcv_[i] = f.lcv_[i]->copy();
    }

    // destructor
    ~InputFunction() {
        for(InputComponentVector::const_iterator ilcv
                = lcv_.begin(); ilcv != lcv_.end(); ++ilcv)
            delete *ilcv;
    }

    // number of observables considered.
    // Not all of these observables have to be claimed by input components.
    // Accordingly, not all observables must enter the input function.
    int nobservables() const { return nobs_; }

    // number of input components
    int ncomponents() const { return lcv_.size(); }

    // access to input components
    const InputComponent& component(int i) const 
    { assert(i >= 0 && i < ncomponents()); return *(lcv_[i]); }
    InputComponent& component(int i)
    { assert(i >= 0 && i < ncomponents()); return *(lcv_[i]); }

    // get/set the constraint penalty factor
    double constraint_penalty_factor() const
    { return constraint_penalty_factor_; }
    void constraint_penalty_factor(double d)
    { constraint_penalty_factor_ = d; }

    // add input component
    int add(const InputComponent& ic);

    // add constraint
    int add_constraint(const InputComponent& c) {
        int result;
        if((result = add(c)) >= 0)
            is_constraint_[is_constraint_.size()-1] = true;
        return result;
    }

    // get/set default central values from likelihood components
    const ObservableVector& central_values() const { return central_values_; }
    void central_values(const ObservableVector& cvals)
    { assert(int(cvals.size()) == nobservables()); central_values_ = cvals; }
    double central_value(int i) const 
    { assert(i >= 0 && i < nobservables()); return central_values_[i]; }
    void central_value(int i, double d)
    { assert(i >= 0 && i < nobservables()); central_values_[i] = d; }

    // calculate the input function, optionally with constraint penalty added
    double calc(const ObservableVector& pred,
                const ObservableVector& meas,
                bool with_constraints=false);
    double calc(const ObservableVector& pred,
                bool with_constraints=false)
    { return calc(pred, central_values(), with_constraints); }

    // calculate an individual component of the input function
    double calc_component(int i,
                          const ObservableVector& pred,
                          const ObservableVector& meas);
    double calc_component(int i,
                          const ObservableVector& pred)
    { return calc_component(i, pred, central_values()); }

    // calculate only the constraint penalty
    // (not scaled by constraint_penalty_factor())
    double calc_constraint_penalty(const ObservableVector& pred,
                                   const ObservableVector& meas);
    double calc_constraint_penalty(const ObservableVector& pred)
    { return calc_constraint_penalty(pred, central_values()); }

    // calculate contribution of observable iobs to input function
    double calc_contrib(const IndexVector& iv,
                        const ObservableVector& pred,
                        const ObservableVector& meas,
                        bool with_constraints=false);
    double calc_contrib(const IndexVector& iv,
                        const ObservableVector& pred,
                        bool with_constraints=false)
    { return calc_contrib(iv, pred, central_values(), with_constraints); }
    double calc_contrib(int iobs,
                        const ObservableVector& pred,
                        const ObservableVector& meas,
                        bool with_constraints=false)
    {
        return calc_contrib(IndexVector(1, iobs),
                            pred, meas, with_constraints);
    }
    double calc_contrib(int iobs,
                        const ObservableVector& pred,
                        bool with_constraints=false)
    { return calc_contrib(iobs, pred, central_values(), with_constraints); }
    double calc_contrib(int i1, int i2,
                        const ObservableVector& pred,
                        const ObservableVector& meas,
                        bool with_constraints=false)
    {
        assert(i1 >= 0);
        assert(i2 >= i1);
        IndexVector iv(i2-i1+1);
        for(int i=i1; i <= i2; ++i) iv[i-i1] = i;
        return calc_contrib(iv, pred, meas, with_constraints);
    }
    double calc_contrib(int i1, int i2,
                        const ObservableVector& pred,
                        bool with_constraints=false)
    { return calc_contrib(i1, i2, pred, central_values(), with_constraints); }


    // calculate contribution of observable iobs to constraint penalty
    double calc_penalty_contrib(const IndexVector& iv,
                                const ObservableVector& pred,
                                const ObservableVector& meas);
    double calc_penalty_contrib(const IndexVector& iv,
                                const ObservableVector& pred)
    { return calc_penalty_contrib(iv, pred, central_values()); }
    double calc_penalty_contrib(int iobs,
                                const ObservableVector& pred,
                                const ObservableVector& meas)
    { return calc_penalty_contrib(IndexVector(1, iobs), pred, meas); }
    double calc_penalty_contrib(int iobs,
                                const ObservableVector& pred)
    { return calc_penalty_contrib(iobs, pred, central_values()); }
    double calc_penalty_contrib(int i1, int i2,
                                const ObservableVector& pred,
                                const ObservableVector& meas)
    {
        assert(i1 >= 0);
        assert(i2 >= i1);
        IndexVector iv(i2-i1+1);
        for(int i=i1; i <= i2; ++i) iv[i-i1] = i;
        return calc_penalty_contrib(iv, pred, meas);
    }
    double calc_penalty_contrib(int i1, int i2,
                                const ObservableVector& pred)
    { return calc_penalty_contrib(i1, i2, pred, central_values()); }


    // calculates the derivatives of the input function with respect
    // to the predicted observables (first argument of calc()).
    // The constraint penalty is always included.
    int calc_deriv(const ObservableVector& pred,
                   const ObservableVector& meas,
                   ObservableVector& result);
    int calc_deriv(const ObservableVector& pred,
                   ObservableVector& result)
    { return calc_deriv(pred, central_values(), result); }

    // calculate the input function without systematic errors, optionally
    // with constraint penalty added
    double calc_nosystematics(const ObservableVector& pred,
                              const ObservableVector& meas,
                              bool with_constraints=false);
    double calc_nosystematics(const ObservableVector& pred,
                              bool with_constraints=false)
    { return calc_nosystematics(pred, central_values(), with_constraints); }

    // returns the vector of indices of claimed observables.
    const IndexVector& claimed_observables() { return claimed_observables_; }

    // returns the number of claimed observables
    int n_claimed_observables() const { return claimed_observables_.size(); }

    // set hess to the the Hessian of the likelihood function
    // at pred=meas=obs. The Hessian is the quadratic term in the Taylor
    // expansion of calc_nosystematics(obs, obs), considered as a function of
    // obs (without the factor 1/2). Rows and columns corresponding to
    // unclaimed observables will be zero. Constraint penalties are never
    // included here.
    void get_hessian(Matrix& hess, const ObservableVector& obs);
    void get_hessian(Matrix& hess) { get_hessian(hess, central_values()); }
};


}

#endif

