/*
This file is part of myFitter, a C++ library for maximum likelihood fits.
Copyright (C) 2012, 2013  Martin Wiebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MYFITTER_MINIMIZER_HPP
#define MYFITTER_MINIMIZER_HPP

#include "model.hpp"

namespace myfitter {

class ObjectiveFunction {
private:
    int ndim_;

protected:
    void resize_(int n) { ndim_ = n; }

public:
    ObjectiveFunction() : ndim_(0) { }
    ObjectiveFunction(const ObjectiveFunction& f) : ndim_(f.ndim_) { }

    // calculate function value at x
    virtual double operator()(const ParameterVector& x) = 0;

    // calculate gradient at x and store gradient in g
    virtual void get_gradient(const ParameterVector& x, ParameterVector& g) = 0;

    // calculate function value and gradient at x
    virtual void get_fdf(const ParameterVector& x,
                         double& f, ParameterVector& g) = 0;

    int ndim() const { return ndim_; }
};

class BasicMinimizer {
protected:
    double precision_;
    double line_search_precision_;
    int maxiter_;
    int niter_;
    int verbosity_;
    bool keep_hessian_;

    double f_;
    ParameterVector x_; 
    ParameterVector grad_; 
    Matrix invhess_;
    
public:
    BasicMinimizer() :
        precision_(0.01),
        line_search_precision_(0.1),
        maxiter_(100),
        niter_(0),
        verbosity_(0),
        keep_hessian_(false),
        f_(Model::nan())
    { }
    BasicMinimizer(int n) :
        precision_(0.01),
        line_search_precision_(0.1),        
        maxiter_(100),
        niter_(0),
        verbosity_(0),
        keep_hessian_(false),
        f_(Model::nan())
    { resize(n); }
    BasicMinimizer(const BasicMinimizer& m) :
        precision_(m.precision_),
        line_search_precision_(m.line_search_precision_),
        maxiter_(m.maxiter_),
        niter_(m.niter_),
        verbosity_(m.verbosity_),
        keep_hessian_(m.keep_hessian_),
        f_(m.f_),
        x_(m.x_),
        grad_(m.grad_),
        invhess_(m.invhess_)
    { }

    // virtual copy constructor
    // derived classes must overload this!
    virtual BasicMinimizer* copy() const = 0;

    // set number of dimensions
    void resize(int n) {
        x_ = ParameterVector(n, Model::nan());
        grad_ = ParameterVector(n, Model::nan());
        reset_inverse_hessian();
    }

    // get number of dimensions
    int ndim() const { return x_.size(); }

    // get/set the verbosity level
    int verbosity() const { return verbosity_; }
    void verbosity(int i) { verbosity_ = i; }

    // get/set minimization precision
    // minimization is successfully terminated if the norm of the gradient
    // is smaller than tolerance.
    double precision() const { return precision_; }
    void precision(double d) { precision_ = d; }

    // get/set the precision for line searches
    double line_search_precision() const { return line_search_precision_; }
    void line_search_precision(double d) { line_search_precision_ = d; }

    // get/set maximum number of iterations
    int maxiter() const { return maxiter_; }
    void maxiter(int n) { maxiter_ = n; }
    
    // returns number of iterations in last minimization
    int niter() const { return niter_; }

    // returns minimum function value found in last minimization
    double f() const { return f_; }

    // get/set the current point in parameter space
    const ParameterVector& x() const { return x_; }
    double x(int i) const { return x_[i]; }
    void x(const ParameterVector& _x) { x_ = _x; }
    void x(int i, double d) { x_[i] = d; }

    // get the final gradient from the last minimization
    const ParameterVector& gradient() const { return grad_; }

    // get the (estimate of the) inverse hessian from the last minimization
    const Matrix& inverse_hessian() const { return invhess_; }

    // set the inverse hessian to the identity matrix
    void reset_inverse_hessian()
    { invhess_ = boost::numeric::ublas::identity_matrix<double>(ndim()); }

    // if true, the minimizer starts the next minimization with the
    // current inverse hessian
    bool keep_hessian() const { return keep_hessian_; }
    void keep_hessian(bool b) { keep_hessian_ = b; }
    
    // minimization routine
    // derived classes must implement this
    virtual int minimize(ObjectiveFunction& fcn) = 0;
};

class BFGSMinimizer : public BasicMinimizer {
public:
    BFGSMinimizer() : BasicMinimizer() { }
    BFGSMinimizer(int n) : BasicMinimizer(n) { }
    BFGSMinimizer(const BFGSMinimizer& m) : BasicMinimizer(m) { }
    virtual BFGSMinimizer* copy() const { return new BFGSMinimizer(*this); }

    virtual int minimize(ObjectiveFunction& fcn);
};

}

#endif
