#include <iostream>

// include the relevant myfitter headers
#include <myfitter/fitter.hpp>
#include <myfitter/inputcomponents.hpp>

// include our test model
#include "linearmodel.hpp"

// include this if you want to be able to print myfitter::Matrix,
// myfitter::ObservableVector and myfitter::Parameter vector objects with
// std::cout << ...
#include <boost/numeric/ublas/io.hpp>

// everything myFitter-related is in the namespace myfitter, so let's unwrap it.
using namespace myfitter;


int main()
{
    int status;

    // introduce a shorthand for 'LinearModel'
    typedef LinearModel Mod;

    // instantiate the linear model
    Mod fullmodel;

    // Set the typical scale of the parameters.
    // These should be of the same order of magnitude as the errors of the
    // parameters obtained from the fit to the data.
    // It is recommended that you use the public enums of the model class
    // to refer to specific parameter or observable indices.
    fullmodel.scale(Mod::P1, 0.1);
    fullmodel.scale(Mod::P2, 0.1);
    fullmodel.scale(Mod::P3, 0.1);
    fullmodel.scale(Mod::P4, 0.1);

    // Set the ranges in which the parameters should be scanned when calling
    // the model's scan() method. To sample a certain region with a higher
    // density you can set the ranges, call scan(), then reduce the ranges,
    // call scan() again and so forth.
    fullmodel.scan_range(Mod::P1, -2.5, 2.5);
    fullmodel.scan_range(Mod::P2, -2.5, 2.5);
    fullmodel.scan_range(Mod::P3, -2.5, 2.5);
    fullmodel.scan_range(Mod::P4, -2.5, 2.5);


    // Initialise the fitter object.
    // The NPAR and NOBS arguments to the constructor must be bigger or
    // equal to the number of parameters and observables of any model fitted
    // with the fitter object.
    Fitter fitter(Mod::NOBS);

    // increase this number to see more output during the fits
    fitter.minimizer_verbosity(0); 

    // Add input components to the input_function() member of the
    // fitter object. Here we let the observables O1, O2, O4 and O6 have
    // independent Gaussian errors. The signature of the GaussianIC constructor
    // is GaussianIC(parameter index, central value, error)
    fitter.input_function()
        .add(GaussianIC(Mod::O1,  2.27, 0.2));
    fitter.input_function()
        .add(GaussianIC(Mod::O2, -0.66, 0.2));
    fitter.input_function()
        .add(GaussianIC(Mod::O4, -1.36, 0.3));
    fitter.input_function()
        .add(GaussianIC(Mod::O6, +2.79, 0.1));
    
    // To add correlated observables to the input function, we first
    // instantiate the CorrelatedGaussianIC class. Note that the argument to the
    // constructor is the number of observables calculated by the model and
    // _not_ the number of observables contributing to the input component.
    // Observables are added to the input component with the method
    // add(observable index, central value, error). The correlation matrix is
    // initialised with cor(observable1, observable2, correlation). Only
    // off-diagonal elements in one triangle have to be initialised.
    // After initialising the input component, we add it to the input
    // fucntion with fitter.input_function().add(lc)
    CorrelatedGaussianIC ic(Mod::NOBS);
    ic.add(Mod::O3, 0.31, 0.2);
    ic.add(Mod::O5, -0.21, 0.3);
    ic.add(Mod::O7, -2.3 , 0.2);
    ic.cor(Mod::O3, Mod::O5, -0.1);
    ic.cor(Mod::O3, Mod::O7, -0.5);
    ic.cor(Mod::O5, Mod::O7, +0.3);
    fitter.input_function().add(ic);

    
    // Here we build a dictionary of 1000 randomly selected parameter vectors
    // and the corresponding observable vectors and then fit the model. The
    // dictionary is stored in the fullmodel object and used by
    // Fitter::global_fit() to determine the best starting point for the
    // chi-square minimisation.
    
    fullmodel.scan(1000);
    std::cout << "fitting full model..." << std::flush;
    if((status = fitter.global_fit(fullmodel))) {
        std::cout << "failed with exit status " << status << std::endl;
        return 1;
    } else
        std::cout << "done.\n" << std::endl;
    std::cout << fullmodel << std::endl;


    // make a copy of the full model and fix two parameters to obtain the
    // constrained model
    LinearModel constrainedmodel = fullmodel;
    constrainedmodel.parameter(Mod::P1, -0.7); constrainedmodel.fix(Mod::P1);
    constrainedmodel.parameter(Mod::P2, -2.0); constrainedmodel.fix(Mod::P2);

    // clear the dictionary, scan again and fit the constrained model
    constrainedmodel.clear();
    constrainedmodel.scan(1000);
    constrainedmodel.scan(1000);
    std::cout << "\nfitting full model..." << std::flush;
    if((status = fitter.global_fit(constrainedmodel))) {
        std::cout << "failed with exit status " << status << std::endl;
        return 1;
    } else
        std::cout << "done.\n" << std::endl;
    std::cout << constrainedmodel << std::endl;


    // save delta-chisquare for later comparison
    double delta_chisquare
        = constrainedmodel.chisquare() - fullmodel.chisquare();

    // Set the desired relative precision, the number of shots per iteration
    // and the maximum number of iterations for p-value integrations.
    fitter.dvegas_precision(0.03);
    fitter.dvegas_nfirstshots(1000);
    fitter.dvegas_nshots(500);
    fitter.dvegas_niterations(20);

    // increase this number to see more output during the integration
    fitter.dvegas_verbosity(0);

    // Perform likelihood ratio test for nested models and compute p-value.
    // The p-value is the probability that a random vector of toy observables
    // leads to a larger Delta chi-square than the observed data. 
    // As 'observed Delta chi-square' the Fitter::calc_nested_lrt_pvalue()
    // and Fitter::calc_lrt_pvalue() methods take the difference
    // of the chisquare() members of the two model objects passed to them.
    std::cout << "computing p-value for nested models..." << std::flush;
    fitter.calc_nested_lrt_pvalue(fullmodel, constrainedmodel);    
    if(!fitter.reached_precision_goal()) {
        std::cout << "failed to reach desired precision." << std::endl;
        return 1;
    } else if(fitter.failed_shot_ratio() > 0.001) {
        std::cout << "more than 0.1% of shots failed." << std::endl;
        return 1;
    } else
        std::cout << "done." << std::endl;

    std::cout << "\np-value (numerical result):  " << fitter.pvalue()
              << " +/- " << fitter.pvalue_error()
              << std::endl;

    // To get the analytical result, we evaluate the normalised lower incomplete
    // Gamma function with a number of degrees of freedom of 2 (because two
    // parameters were fixed in the constrained model
    int ndof = 2;
    std::cout << "p-value (analytical result): "
              << 1. - boost::math::gamma_p(0.5*ndof,0.5*delta_chisquare)
              << std::endl;

    // Now we perform the likelihood ratio test for non-nested models.  This
    // algorithm is less efficient but also works if the two models are
    // unrelated. (Of course, the two model objects must have the same number of
    // observables and you have to make sure that the observables appear in the
    // same order in the observable vector.) Note that the last p-value
    // integration has changed the state and in particular the chisquare()
    // members of the two model objects. To do the test with the same 'observed'
    // Delta chi-square as before we just fit them again.

    fitter.global_fit(fullmodel);
    fitter.global_fit(constrainedmodel);

    std::cout << "\n\ncomputing p-value for non-nested models..." << std::flush;
    fitter.calc_lrt_pvalue(fullmodel, constrainedmodel);    
    if(!fitter.reached_precision_goal()) {
        std::cout << "failed to reach desired precision." << std::endl;
        return 1;
    } else if(fitter.failed_shot_ratio() > 0.001) {
        std::cout << "more than 0.1% of shots failed." << std::endl;
        return 1;
    } else
        std::cout << "done." << std::endl;

    std::cout << "\np-value (numerical result):  " << fitter.pvalue()
              << " +/- " << fitter.pvalue_error()
              << std::endl;

    return 0;
}
