#include <iostream>
#include <myfitter/model.hpp>


class LinearModel : public myfitter::Model {
private:
    // this model implements an affine-linear relation between four parameters
    // and seven observables: obs = mat_ * par + vec_
    myfitter::Matrix mat_;
    myfitter::ObservableVector vec_;

public:
    // It is not required but highly recommended that you define public enums
    // in you model class, which give intuitive names to the parameter and
    // observable indices in your model. Even though the number of parameters
    // and observables is accessible dynamically via the nparameters() and
    // nobservables() methods, it is also recommended that you provide static
    // information about the number of parameters and observables by
    // defining or overloading the static constants NPAR and NOBS

    // parameters
    enum {P1, P2, P3, P4};
    static const int NPAR = P4+1;

    // observables
    enum {O1, O2, O3, O4, O5, O6, O7};
    static const int NOBS = O7+1;

    // constructors and assignment operators:
    // Each model should define a default constructor and a copy constructor
    // and overload the virtual assignment operator.

    // default constructor:
    // calls the protected constructor Model(int, int) to initialise
    // the (dynamical) information about the number of parameters and
    // observables. It also initialises the mat_ and vec_ members.
    LinearModel() : myfitter::Model(NPAR,NOBS), mat_(NOBS,NPAR), vec_(NOBS)
    {
        mat_(O1, P1) = +0.3;
        mat_(O1, P2) = -1.2;
        mat_(O1, P3) = +0.0;
        mat_(O1, P4) = -0.7;

        mat_(O2, P1) = +0.4;
        mat_(O2, P2) = +0.6;
        mat_(O2, P3) = -0.9;
        mat_(O2, P4) = +1.0;

        mat_(O3, P1) = -1.5;
        mat_(O3, P2) = -0.1;
        mat_(O3, P3) = -0.1;
        mat_(O3, P4) = -0.5;

        mat_(O4, P1) = +0.0;
        mat_(O4, P2) = +0.0;
        mat_(O4, P3) = -0.6;
        mat_(O4, P4) = +0.2;

        mat_(O5, P1) = +0.4;
        mat_(O5, P2) = -0.4;
        mat_(O5, P3) = -1.1;
        mat_(O5, P4) = +0.7;

        mat_(O6, P1) = -0.3;
        mat_(O6, P2) = -0.6;
        mat_(O6, P3) = +0.2;
        mat_(O6, P4) = +0.9;

        mat_(O7, P1) = -0.1;
        mat_(O7, P2) = +0.9;
        mat_(O7, P3) = -1.0;
        mat_(O7, P4) = +0.0;

        vec_(O1) = +0.5;
        vec_(O2) = +0.5;
        vec_(O3) = -0.4;
        vec_(O4) = -0.8;
        vec_(O5) = -0.9;
        vec_(O6) = +0.2;
        vec_(O7) = +0.0;
    }

    // copy constructor:
    // calls the copy constructor of the base class and the members mat_ and
    // vec_
    LinearModel(const LinearModel& m) :
        myfitter::Model(m), mat_(m.mat_), vec_(m.vec_) { }

    // virtual assignment operator:
    // calls the assignment operators of the base class and the members
    // mat_ and vec_.
    virtual const LinearModel& operator=(const LinearModel& m) {
        myfitter::Model::operator=(m);
        mat_ = m.mat_;
        vec_ = m.vec_;
        return *this;
    }

    // calc method:
    // sets the elements of the protected member observables_ based on the
    // current values of the parameters (parameters() returns the parameter
    // vector). 
    virtual int calc() {
        observables_
            .assign(boost::numeric::ublas::prod(mat_, parameters()) + vec_);
        // If possible, call the calc() method of the base class somewhere
        // in here. If you don't, make sure you set the public member
        // needs_update_ to false to avoid repeated evaluation with the
        // same parameters. The return value should be zero if the evaluation
        // was successful and nonzero otherwise.
        return myfitter::Model::calc();
    }

    // calc_deriv method:
    // fills the protected matrix derivatives_ with the derivatives of all
    // observables with respect to all parameters.  The default implementation
    // in Model calculates all derivatives numerically. You may overload this
    // method to implement e.g. analytical formulae for the derivatives. If you
    // do not derive directly from Model you probably have to overload this. You
    // can call the protected method numerical_derivative_(int i) to numerically
    // calculate the derivatives of all observables with respect to the
    // parameter i. The results are put in the corresponding elements of
    // derivatives_. The return value should be zero if the evaluation was
    // successful and nonzero otherwise.
    virtual int calc_deriv() {
        derivatives_.assign(boost::numeric::ublas::trans(mat_));
        return 0;
    }
};

// overload operator<< to print information about parameters, observables
// and chi-square value
std::ostream& operator<<(std::ostream& os, const LinearModel& mod)
{
    return os << "P1           = " << mod.parameter(LinearModel::P1)
              << "\nP2           = " << mod.parameter(LinearModel::P2)
              << "\nP3           = " << mod.parameter(LinearModel::P3)
              << "\nP4           = " << mod.parameter(LinearModel::P4)
              << "\n"
              << "\nO1           = " << mod.observable(LinearModel::O1)
              << "\nO2           = " << mod.observable(LinearModel::O2)
              << "\nO3           = " << mod.observable(LinearModel::O3)
              << "\nO4           = " << mod.observable(LinearModel::O4)
              << "\nO5           = " << mod.observable(LinearModel::O5)
              << "\nO6           = " << mod.observable(LinearModel::O6)
              << "\nO7           = " << mod.observable(LinearModel::O7)
              << "\n"
              << "\nchi-square   = " << mod.chisquare()
              << std::endl;
}
